pro reconstruct_aver_drift, drifts_save, npairs_save, file_scalars=file_scalars, $
                            ind_scans=ind_scans, ind_subscans=ind_subscans, $
                            noise_level=noise_level, time_1=time_1, t_proc=t_proc


;;; computes absolute average drift as a function of time, from drift differences matrix
;;; solution not unique, containing an oscillatory component removed in subtract_aver_drift.pro

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; copyright 2007-2010 H\'el\`ene Roussel
;;; roussel@iap.fr
;;; Institut d'Astrophysique de Paris - Universit\'e Pierre et Marie Curie, France

;;; download and documentation: http://www2.iap.fr/users/roussel/herschel/

;;; Scanamorphos is distributed under the terms of the GNU General Public License
;;; (http://www.gnu.org/licenses/gpl.html).

;;; July 2010: public release version 1

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~



restore, file_scalars

restore, file_diff_drifts

restore, file_times

restore, file_time_coarse

flag_nza_coarse = interpol(flag_nza, time_s, time_save)

cwmin = 3


diff_drifts(lindgen(nt_save), lindgen(nt_save)) = 0.

if keyword_set(visu) then begin
   time_2 = systime(/seconds)
   t_proc += (time_2 - time_1)
   disp_ima, win=5, diff_drifts, min_map=-noise_level/3., max_map=noise_level/3., $
      title='matrix of average drift differences', col_tab=15
   time_1 = systime(/seconds)
endif



drifts_save = fltarr(nt_save)
npairs_save = fltarr(nt_save)
drifts = fltarr(nt_save)
npairs = fltarr(nt_save)


iter_rec = 0

repeat begin
   w = where(npairs_save gt 0., ntot_prev)
   drifts_save_prev = drifts_save
   npairs_save_prev = npairs_save

   scan_st = nscans - 1 - (iter_rec mod nscans)
   w = where(ind_subscans ge ind_scans(scan_st) and ind_subscans le ind_scans(scan_st+1) - 1, cw)
   if cw le 2 then subs_st = w(cw-1) else subs_st = w(cw-2) - ((iter_rec / 2) mod (cw - 2))


   n_t2 = 0
   ti = time_s(ind_subscans(subs_st))
   tf = time_s(ind_subscans(subs_st+1)-1)
   ind_save = where(time_save ge 2.*ti/3. + tf/3. and time_save le 2.*tf/3. + ti/3., c_save)
   for t1 = ind_save(0), ind_save(c_save-1) do begin
      slice_npairs = diff_drifts(t1:*, t1)
      w = where(slice_npairs gt 0., c2)
      if c2 gt n_t2 then begin
         n_t2 = c2
         t1_start = t1
      endif
   endfor


   it = t1_start - 1

   while it gt 0 do begin
      slice_npairs = [reform(diff_drifts(it, 0:it-1)), -9., diff_drifts(it+1:*, it)]
      t2 = where(slice_npairs gt 0., c2)
      if c2 gt 0 then begin
         drifts *= 0.
         npairs *= 0.
         w = where(t2 lt it, cw)
         if cw ge 1 then begin
            drifts(t2(w)) = diff_drifts(t2(w), it) + drifts_save(it)
            npairs(t2(w)) = diff_drifts(it, t2(w))
         endif
         w = where(t2 gt it, cw)
         if cw ge 1 then begin
            drifts(t2(w)) = drifts_save(it) - diff_drifts(it, t2(w))
            npairs(t2(w)) = diff_drifts(t2(w), it)
         endif
         wnz = where(npairs gt 0., cnz)
         if max(npairs_save) eq 0. then w = where(npairs gt 0., cw) $
            else w = where(npairs_save * npairs gt 0., cw)
         if cw ge cwmin then begin
            sum_npairs = npairs(w) + npairs_save(w)
            off = total((drifts(w) - drifts_save(w)) * npairs(w)) / total(npairs(w))
            dev = sqrt(mean((drifts(w) - drifts_save(w) - off)^2.))
            wkeep = where(abs(drifts(w) - drifts_save(w) - off) le 2.5 * dev, ckeep, complement=wz, ncomplement=cz)
            if ckeep ge 1 then off = total((drifts(w(wkeep)) - drifts_save(w(wkeep))) * npairs(w(wkeep))) $
               / total(npairs(w(wkeep)))
            if cz ge 1 then npairs(w(wz)) = 0.
            drifts -= off
            sum_npairs = npairs + npairs_save
            drifts_save = drifts_save * npairs_save + drifts * npairs
            wpos = where(sum_npairs gt 0.)
            drifts_save(wpos) /= sum_npairs(wpos)
            npairs_save = sum_npairs
         endif
      endif
      it -= 1
   endwhile


   it = t1_start

   while it lt nt_save - 1 do begin
      slice_npairs = [reform(diff_drifts(it, 0:it-1)), -9., diff_drifts(it+1:*, it)]
      t2 = where(slice_npairs gt 0., c2)
      if c2 gt 0 then begin
         drifts *= 0.
         npairs *= 0.
         w = where(t2 lt it, cw)
         if cw ge 1 then begin
            drifts(t2(w)) = diff_drifts(t2(w), it) + drifts_save(it)
            npairs(t2(w)) = diff_drifts(it, t2(w))
         endif
         w = where(t2 gt it, cw)
         if cw ge 1 then begin
            drifts(t2(w)) = drifts_save(it) - diff_drifts(it, t2(w))
            npairs(t2(w)) = diff_drifts(t2(w), it)
         endif
         wnz = where(npairs gt 0., cnz)
         if max(npairs_save) eq 0. then w = where(npairs gt 0., cw) $
            else w = where(npairs_save * npairs gt 0., cw)
         if cw ge cwmin then begin
            sum_npairs = npairs(w) + npairs_save(w)
            off = total((drifts(w) - drifts_save(w)) * npairs(w)) / total(npairs(w))
            dev = sqrt(mean((drifts(w) - drifts_save(w) - off)^2.))
            wkeep = where(abs(drifts(w) - drifts_save(w) - off) le 2.5 * dev, ckeep, complement=wz, ncomplement=cz)
            if ckeep ge 1 then off = total((drifts(w(wkeep)) - drifts_save(w(wkeep))) * npairs(w(wkeep))) $
               / total(npairs(w(wkeep)))
            if cz ge 1 then npairs(w(wz)) = 0.
            drifts -= off
            sum_npairs = npairs + npairs_save
            drifts_save = drifts_save * npairs_save + drifts * npairs
            wpos = where(sum_npairs gt 0.)
            drifts_save(wpos) /= sum_npairs(wpos)
            npairs_save = sum_npairs
         endif
      endif
      it += 1
   endwhile


   iter_rec += 1

   w = where(npairs_save gt 0., ntot)
   resi = max(abs(drifts_save_prev(w) - drifts_save(w)))
endrep until ((resi lt 0.05 * noise_level and ntot le ntot_prev and iter_rec ge 6) or iter_rec ge 24)


end
